from flask import render_template, flash, Markup, send_from_directory
from flask_jwt_extended import jwt_required, get_current_user

from . import main

from seamm_dashboard.routes.api.auth import refresh_expiring_jwts


@main.route("/")
@jwt_required(optional=True)
def index():
    """Homepage"""
    if get_current_user() is None:
        flash(
            Markup(
                "You are currently viewing the dashboard as a public user."
                + " <a href='login'>Log In</a> to see your jobs."
            ),
            category="public",
        )
    return render_template("index.html")


@main.route("/401")
def unauthorized():
    """Unauthorized Access"""
    return render_template("401.html")


@main.route("/views/id/<id>")
@main.route("/views//id/<id>")
def get_sample(id):
    return "<h3> this is id: " + id + "</h3>"


@main.route("/views/<path:path>")
@main.route("/views//<path:path>")
@jwt_required(optional=True)
def send_view(path):
    return render_template("views/" + path)


@main.route("/static/<path:path>")
def send_js(path):
    return send_from_directory("static", path)


@main.after_request
def main_refresh(response):
    return refresh_expiring_jwts(response)



"""
API calls for flowcharts
"""

from seamm_datastore.database.models import Flowchart
from seamm_datastore.database.schema import FlowchartSchema
from flask import Response
from flask_jwt_extended import jwt_required

from seamm_dashboard import authorize

import json

__all__ = ["get_flowcharts", "get_flowchart", "get_cytoscape", "add_flowchart"]


def add_flowchart():
    pass


@jwt_required(optional=True)
def get_flowcharts(description=None, limit=None):

    # If limit is not set, set limit to all jobs in DB.
    if limit is None:
        limit = Flowchart.query.count()

    if description is not None:
        flowcharts = Flowchart.query.filter(
            Flowchart.description.contains(description)
        ).limit(limit)
    else:
        flowcharts = Flowchart.query.limit(limit)

    authorized_flowcharts = []
    for flowchart in flowcharts:
        if authorize.read(flowchart):
            authorized_flowcharts.append(flowchart)
        else:
            for project in flowchart.projects:
                if authorize.read(project):
                    authorized_flowcharts.append(flowchart)
                    break

    flowcharts_schema = FlowchartSchema(many=True)

    return flowcharts_schema.dump(authorized_flowcharts), 200


@jwt_required(optional=True)
def get_flowchart(id):
    """
    Function for api endpoint api/flowcharts/{id}

    Parameters
    ----------
    id : the ID of the flowchart to return
    """
    flowchart = Flowchart.query.get(id)

    if flowchart is None:
        return Response(status=404)

    authorized = False
    if authorize.read(flowchart):
        authorized = True

    if not authorized:
        for project in flowchart.projects:
            if authorize.read(project):
                authorized = True
                break

    if not authorized:
        return Response(status=401)

    flowchart_schema = FlowchartSchema(many=False)
    return flowchart_schema.dump(flowchart), 200


@jwt_required(optional=True)
def get_cytoscape(id, flowchartKeys=None):
    """
    Function for getting cytoscape elements for a flowchart.
    """

    flowchart = Flowchart.query.get(id)

    if flowchart is None:
        return Response(status=404)

    authorized = False

    if authorize.read(flowchart):
        authorized = True
    if not authorized:
        for project in flowchart.projects:
            if authorize.read(project):
                authorized = True
                break

    if not authorized:
        return Response(status=401)

    important_stuff = {}
    important_stuff = flowchart.json

    if isinstance(important_stuff, str):
        important_stuff = json.loads(flowchart.json)

    elements = []

    for node_number, node in enumerate(important_stuff["nodes"]):
        url = "#"
        # Build elements for cytoscape
        elements.append(
            {
                "data": {
                    "id": node["attributes"]["_uuid"],
                    "name": node["attributes"]["_title"],
                    "url": url,
                },
                "position": {
                    "x": node["attributes"]["x"],
                    "y": node["attributes"]["y"],
                },
                "description": "",
            }
        )

    for edge in important_stuff["edges"]:
        node1_id = edge["node1"]
        node2_id = edge["node2"]
        edge_data = {
            "data": {
                "id": str(node1_id) + "_" + str(node2_id),
                "source": node1_id,
                "target": node2_id,
            },
        }

        elements.append(edge_data)
    return elements, 201



"""
Routes for REST authentication
"""

from datetime import timedelta, datetime, timezone

from flask import jsonify, Response, make_response, request, redirect, url_for
from flask_jwt_extended import (
    create_access_token,
    create_refresh_token,
    set_access_cookies,
    set_refresh_cookies,
    get_jwt_identity,
    unset_jwt_cookies,
    get_jwt,
)

from seamm_datastore.database.models import User
from seamm_datastore.database.schema import UserSchema

from seamm_dashboard import jwt

__all__ = ["get_auth_token", "remove_auth_token", "refresh_expiring_jwts"]


def refresh_expiring_jwts(response):
    """This will automatically refresh tokens that are within 30 minutes of expiring"""

    try:
        exp_timestamp = get_jwt()["exp"]
        now = datetime.now(timezone.utc)
        target_timestamp = datetime.timestamp(now + timedelta(seconds=0))
        if target_timestamp > exp_timestamp:
            access_token = create_access_token(identity=get_jwt_identity())
            set_access_cookies(response, access_token)
        return response
    except (RuntimeError, KeyError):
        # Case where there is not a valid JWT. Just return the original respone
        return response

    return response


def create_tokens(user):
    """
    Create a token for a given user object. Not an api endpoint
    """

    user_schema = UserSchema(many=False)
    user = user_schema.dump(user)
    access_token = create_access_token(identity=user)
    refresh_token = create_refresh_token(identity=user)

    return access_token, refresh_token


def get_auth_token(body):
    """
    Endpoint for api/auth/token
    """

    username = body.pop("username")
    password = body.pop("password")

    user = User.query.filter_by(username=username).one_or_none()

    if not user:
        return jsonify({"msg": f"User {username} not found."}), 400

    if user.verify_password(password):

        access_token, refresh_token = create_tokens(user)

        resp = Response({"login": True})
        set_access_cookies(resp, access_token)
        set_refresh_cookies(resp, refresh_token)
        return resp, 200
    else:
        return jsonify({"msg": "Incorrect password."}), 400


def remove_auth_token():
    resp = make_response({"logout": True})
    unset_jwt_cookies(resp)
    return resp, 200


@jwt.unauthorized_loader
def needed_token_callback(_):
    if "api" in request.url:
        return (
            jsonify(
                {
                    "status": 401,
                    "sub_status": 42,
                    "msg": "You are missing a token, please login.",
                }
            ),
            401,
        )

    else:
        return redirect(url_for("auth.login"))


@jwt.expired_token_loader
def my_expired_token_callback(jwt_header, expired_token):
    user_agent = request.environ.get("HTTP_USER_AGENT", [])

    # If we are in a browser we want to redirect to the log out page
    # If we are not in a browser, we will return a json telling them that
    # the current token is expired.

    # This check should catch all Mozilla compatible browsers.
    # Ref: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/User-Agent/Firefox
    browser = any(["Mozilla" in user_agent])

    if browser:
        return redirect(url_for("auth.logout", expired=True))

    token_type = expired_token["type"]
    return (
        jsonify(
            {
                "status": 401,
                "sub_status": 42,
                "msg": "The {} token has expired.".format(token_type)
                + "If you are in a browser and seeing this message, "
                + "please clear your browser cookies.",
            }
        ),
        401,
    )



import json

import requests

from flask import request, render_template, flash, redirect, url_for

from flask_jwt_extended import jwt_required, get_current_user

from seamm_datastore.database.models import Project, Job

from seamm_dashboard import db, authorize

from . import jobs

from seamm_dashboard.routes.jobs.forms import EditJob, ImportJob
from seamm_dashboard.routes.api.auth import refresh_expiring_jwts


@jobs.route("/views/jobs/")
@jobs.route("/views//jobs/")
@jwt_required(optional=True)
def jobs_list():
    return render_template("jobs/jobs_list.html", project=False)


@jobs.route("/views/jobs/<id>")
@jobs.route("/views//jobs/<id>")
@jwt_required(optional=True)
def job_details(id):

    from seamm_datastore.util import NotAuthorizedError

    try:
        job = Job.get_by_id(id)
    except NotAuthorizedError:
        return render_template("401.html")

    edit_job = authorize.update(job)

    own_string = "You do not own this job."

    if job.owner == get_current_user():
        own_string = "You are the job owner."

    # Build the url ourselves.
    base_url = url_for("main.index")
    edit_url = base_url + f"jobs/{id}/edit"

    # Figure out if we can use cdn for plotly
    plotly_location = "https://cdn.plot.ly/plotly-2.9.0.min.js"
    found_plotly = requests.get(f"{plotly_location}").status_code == 200

    if not found_plotly:
        plotly_location = url_for(
            "static", filename="node_modules/plotly.js-dist-min/plotly.min.js"
        )

    return render_template(
        "jobs/job_report.html",
        edit_job=edit_job,
        job=job,
        edit_url=edit_url,
        own_string=own_string,
        plotly=plotly_location,
    )


@jobs.route("/jobs/<job_id>/edit", methods=["GET", "POST"])
@jwt_required(optional=True)
def edit_job(job_id):

    job = Job.query.get(job_id)

    if not authorize.update(job):
        return render_template("401.html")

    form = EditJob()

    # Build the url ourselves.
    base_url = url_for("main.index")
    job_url = base_url + f"#jobs/{job_id}"

    if form.validate_on_submit():
        job.title = form.name.data
        job.description = form.notes.data
        db.session.commit()
        flash("Job updated successfully.", "successs")

        return redirect(job_url)

    elif request.method == "GET":
        form.name.data = job.title
        form.notes.data = job.description

    return render_template(
        "jobs/edit_job.html", title=f"Edit Job {job_id}", form=form, back_url=job_url
    )


@jobs.route("/jobs/import", methods=["GET", "POST"])
@jwt_required()
def import_job():
    # TODO - rewrite
    form = ImportJob()

    if form.validate_on_submit():
        data = json.load(form.outfile.data.stream)

        working_directory = data["working directory"]

        job_info = Job.parse_job_data(working_directory)

        if form.title.data:
            job_info["title"] = form.title.data

        projects = data["projects"]

        project_objs = []
        for project in projects:
            query_project = Project.query.filter_by(name=project).one_or_none()

            # Create project if it doesn't exist
            if not query_project:
                query_project = Project(name=project)
                db.session.add(query_project)
                db.session.commit()

            project_objs.append(query_project)

        # TODO Add job
        # add_job(working_directory, job_info["title"], project_objs)
        flash(f"Job {job_info['id']} successfully added to database.")
        return redirect(url_for("main.index"))

    return render_template("jobs/import_job.html", form=form)


@jobs.after_request
def job_refresh(response):
    return refresh_expiring_jwts(response)



from flask import render_template
from flask_jwt_extended import jwt_required

from . import flowcharts

from seamm_dashboard.routes.api.auth import refresh_expiring_jwts


@flowcharts.route("/views/flowcharts")
@jwt_required(optional=True)
def flowchart_list():
    return render_template("flowcharts/flowchart_list.html")


@flowcharts.route("/views/flowcharts/<id>")
@flowcharts.route("/views/flowcharts/<id>/<flowchart_keys>")
@jwt_required(optional=True)
def flowchart_details(id):
    return render_template("flowcharts/render_flowchart.html")


@flowcharts.after_request
def job_refresh(response):
    return refresh_expiring_jwts(response)



import pytest

import os
import shutil
from datetime import datetime

from seamm_dashboard import create_app, db

from seamm_datastore.database.models import (
    Job,
    Flowchart,
    Project,
    User,
    Role,
    UserProjectAssociation,
    Group,
)
from selenium import webdriver

# Adds chromedriver binary to path:
import chromedriver_binary  # noqa: F401


def _get_cookie_from_response(response, cookie_name):
    """
    Function grabbed from testing suite of flask-jwt-extended.
    """
    cookie_headers = response.headers.getlist("Set-Cookie")

    for header in cookie_headers:
        attributes = header.split(";")
        if cookie_name in attributes[0]:
            cookie = {}
            for attr in attributes:
                split = attr.split("=")
                cookie[split[0].strip().lower()] = split[1] if len(split) > 1 else True
            return cookie
    return None


@pytest.fixture(scope="session")
def project_directory(tmpdir_factory):

    # Copy our project files to a tmpdir
    dir_path = os.path.dirname(os.path.realpath(__file__))
    real_projects_path = os.path.realpath(
        os.path.join(dir_path, "..", "..", "data", "projects")
    )
    real_MyProject_path = os.path.join(real_projects_path, "MyProject")

    temp_projects_path = tmpdir_factory.mktemp("projects")
    myproject_path = str(temp_projects_path.mkdir("MyProject"))

    return_path = shutil.copytree(
        real_MyProject_path, myproject_path, dirs_exist_ok=True
    )

    job_lock = os.path.join(temp_projects_path, "job.id")

    with open(job_lock, "w") as f:
        f.write("!MolSSI job_id 1.0\n11")

    return return_path


@pytest.fixture(scope="module")
def app(project_directory):
    import seamm_dashboard

    seamm_dashboard.datastore = os.path.split(project_directory)[0]

    test_project_path = project_directory

    flask_app = create_app("testing")
    app_context = flask_app.app_context()
    app_context.push()

    # Create a sample project
    test_project = {
        "name": "MyProject",
        "path": test_project_path,
        "owner_id": 3,
        "id": 100,
    }

    project = Project(**test_project)

    # Create some sample roles
    admin_role = Role.query.filter_by(name="admin").one_or_none()
    manager_role = Role.query.filter_by(name="group manager").one_or_none()
    user_role = Role.query.filter_by(name="user").one_or_none()

    # Get groups for the users
    groups = Group.query.all()

    # Create a sample user.
    test_user = User(
        username="sample_user",
        password="sample_password",
        roles=[user_role],
        groups=[groups[1]],
    )
    test_admin = User(
        username="admin_user", password="iamadmin", roles=[admin_role], groups=groups
    )

    # Fill in some data
    sub_time = datetime.fromisoformat("2016-08-29T09:12:33.001000+00:00")
    # "submitted": parser.parse("2016-08-29T09:12:33.001000+00:00"),
    job1_data = {
        "flowchart_id": 100,
        "id": 1,
        "path": os.path.realpath(os.path.join(test_project_path, "Job_000001")),
        "submitted": sub_time,
        "projects": [project],
        "owner_id": 3,
        "status": "finished",
    }

    # Fill in some data
    job2_data = {
        "flowchart_id": 100,
        "id": 2,
        "path": os.path.realpath(os.path.join(test_project_path, "Job_000002")),
        "submitted": sub_time,
        "projects": [project],
        "owner_id": 3,
        "status": "finished",
    }

    # More data - this job path (probably) doesn't actually exist
    job3_data = {
        "flowchart_id": 100,
        "id": 3,
        "path": "/Users/username/seamm/projects",
        "submitted": sub_time,
        "projects": [project],
        "owner_id": 3,
        "status": "finished",
    }

    # Load a simple flowchart
    current_location = os.path.dirname(os.path.realpath(__file__))
    flowchart_data = Flowchart.parse_flowchart_file(
        os.path.join(current_location, "..", "..", "data", "sample.flow")
    )
    # Make the ID easier
    flowchart_data[0]["id"] = 100
    flowchart_data[0]["owner_id"] = 3
    flowchart_data[0]["json"] = flowchart_data[1]

    # Save the fake data to the db
    job1 = Job(**job1_data)
    job2 = Job(**job2_data)

    # Add job3 and readable for world.
    job3 = Job(**job3_data)
    job3.permissions = {"other": ["read"]}

    # Add visitor and give read access to a project
    visitor = User(username="visitor", password="visitor", id=10)
    a = UserProjectAssociation(
        permissions=["read"], resource_id=project.id, entity_id=visitor.id
    )
    a.job = job1
    visitor.special_projects.append(a)
    db.session.add(a)
    db.session.add(visitor)

    flowchart = Flowchart(**flowchart_data[0])
    db.session.add(test_user)
    db.session.add(admin_role)
    db.session.add(manager_role)
    db.session.add(test_admin)
    db.session.add(project)
    db.session.add(job1)
    db.session.add(job2)
    db.session.add(job3)
    db.session.add(flowchart)
    db.session.commit()

    yield flask_app

    # clean up
    app_context.pop()


@pytest.fixture(scope="module")
def client(app):

    my_client = app.test_client()
    yield my_client


@pytest.fixture(scope="module")
def auth_client(client):
    auth_client = client
    response = auth_client.post(
        "api/auth/token",
        json=dict(
            username="sample_user",
            password="sample_password",
        ),
        follow_redirects=True,
    )

    csrf_token = _get_cookie_from_response(response, "csrf_access_token")[
        "csrf_access_token"
    ]

    yield auth_client, csrf_token

    auth_client.get("api/auth/token/remove", follow_redirects=True)


@pytest.fixture(scope="module")
def visitor_client(client):
    visitor_client = client
    response = visitor_client.post(
        "api/auth/token",
        json=dict(
            username="visitor",
            password="visitor",
        ),
        follow_redirects=True,
    )

    csrf_token = _get_cookie_from_response(response, "csrf_access_token")[
        "csrf_access_token"
    ]

    yield visitor_client, csrf_token

    visitor_client.get("api/auth/token/remove", follow_redirects=True)


@pytest.fixture(scope="module")
def admin_client(app):
    client = app.test_client()
    response = client.post(
        "api/auth/token",
        json=dict(
            username="admin_user",
            password="iamadmin",
        ),
        follow_redirects=True,
    )

    csrf_token = _get_cookie_from_response(response, "csrf_access_token")[
        "csrf_access_token"
    ]

    yield client, csrf_token

    client.get("api/auth/token/remove", follow_redirects=True)


@pytest.fixture
def chrome_driver():
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument("--headless")
    chrome_options.add_argument("--no-sandbox")
    # executable_path = os.getenv('EXECUTABLE_PATH')
    driver = webdriver.Chrome(options=chrome_options)

    yield driver

    driver.close()




# This file was generated by 'versioneer.py' (0.18) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

import json

version_json = '''
{
 "date": "2022-03-18T10:28:01-0400",
 "dirty": false,
 "error": null,
 "full-revisionid": "71e43c5989e051e0d3bac667af4a94a2d6215d7c",
 "version": "2022.3.18"
}
'''  # END VERSION_JSON


def get_versions():
    return json.loads(version_json)



"""
API calls for users (creating, logging in, logging out)

"""

from flask import Response
from flask_jwt_extended import jwt_required

from seamm_dashboard import db, authorize
from seamm_datastore.database.models import User, Role, Group

from seamm_datastore.database.schema import UserSchema, RoleSchema, GroupSchema

__all__ = ["add_user", "get_users"]


def _process_user_body(request_data, original_user_data=None):
    """This function is a private internal function to process user data to get it in a
    format which can be added to the database.

    Takes request data in json and converts to user object. Returns 400 status if not
    valid.
    """

    username = request_data["username"]
    password = request_data["password"]

    if not original_user_data:
        # Check that both username and password are supplied
        if username is None or password is None:
            return Response(
                "Both username and password must be supplied to create new user",
                status=400,
            )
        # Create starting user object with the appropriate information
        user_info = {
            "username": username,
            "password": password,
            "roles": [],
            "groups": [],
        }
        user = User(**user_info)

        # Check if username exists.
        if User.query.filter_by(username=username).first() is not None:
            return Response(
                f"User with username '{username}'' already exists", status=400
            )

    else:
        user = original_user_data
        user.roles = []
        user.groups = []

        # Check if username exists. First we check if the name is in the database. If
        # the name is in the database, this may be an update, so check if the supplied
        # username is equal to the original user info.
        if (
            User.query.filter_by(username=username).first() is not None
            and user.username != username
        ):
            return Response(
                f"User with username '{username}' already exists", status=400
            )

        # After check, update username and password.
        user.username = username

        # only set the password if the password is not blank
        if password:
            user.password = password

    possible_keys = ["roles", "groups", "first_name", "last_name", "email"]

    map_values = {
        "roles": Role,
        "roles_values": [],
        "groups": Group,
        "group_values": [],
    }

    for key in possible_keys:
        try:
            if key == "roles" or key == "groups":
                model = map_values[key]
                for listed_value in request_data[key]:
                    is_model = model.query.filter_by(name=listed_value).first()
                    if not is_model:
                        return Response(
                            f"{listed_value} is not an available value for {key}.",
                            status=400,
                        )
                    else:
                        if getattr(user, key) is None:
                            setattr(user, key, [])
                        getattr(user, key).append(is_model)
            else:
                setattr(user, key, request_data[key])
        except KeyError:
            # Not a key in the body. Pass.
            pass

    return user


@jwt_required(optional=True)
@authorize.has_role("admin")
def add_user(body):

    get_data = _process_user_body(body)

    if isinstance(get_data, Response):
        return get_data

    db.session.add(get_data)
    db.session.commit()

    return get_data.id, 201


@jwt_required(optional=True)
@authorize.has_role("admin")
def get_users():
    users = User.query.all()

    user_list = []
    for current_user in users:
        roles = current_user.roles
        role_schema = RoleSchema(many=True)
        roles = role_schema.dump(roles)

        user_roles = []
        for role in roles:
            user_roles.append(role["name"])

        groups = current_user.groups
        group_schema = GroupSchema(many=True)
        groups = group_schema.dump(groups)

        user_groups = []
        for group in groups:
            user_groups.append(group["name"])

        user_info = UserSchema(many=False).dump(current_user)
        user_info["roles"] = user_roles
        user_info["groups"] = user_groups

        user_list.append(user_info)

    return user_list, 200



import logging
import os
from pathlib import Path
from datetime import timedelta

import connexion

# from flask_admin import Admin
from flask_debugtoolbar import DebugToolbarExtension
from flask_bootstrap import Bootstrap
from flask_cors import CORS
from flask_mail import Mail
from flask_moment import Moment
from flask_sqlalchemy import SQLAlchemy
from flask_marshmallow import Marshmallow
from flask_compress import Compress

from flask_authorize import Authorize

from .jwt_patch import flask_jwt_extended

from .config import config
from .template_filters import replace_empty
from .setup_logging import setup_logging
from .setup_argparsing import options, parser


# Handle versioneer
from ._version import get_versions

__author__ = """Jessica Nash"""
__email__ = "janash@vt.edu"
versions = get_versions()
__version__ = versions["version"]
__git_revision__ = versions["full-revisionid"]
del get_versions, versions

# Ensure that the projects directory exists.
datastore_path = Path(options["datastore"]).expanduser().resolve()
datastore = str(datastore_path)
projects_path = datastore_path / "projects"
projects_path.mkdir(parents=True, exist_ok=True)

# Setup the logging, now that we know where the datastore is
setup_logging(datastore, options)
logger = logging.getLogger("dashboard")

# If there is no database we need to initialize!
db_path = datastore_path / "seamm.db"
if not db_path.exists():
    logger.warning("The database does not exists, so forcing initialization")
    options["initialize"] = True

# Two of the Flask options cannot be reset, and should (apparently) be
# handled with environment variables ... so if they are in the options
# set the correct environment variables. Carefully!

if "env" in options:
    if "FLASK_ENV" in os.environ and options["env"] != os.environ["FLASK_ENV"]:
        logger.warning(
            (
                "The environment variable FLASK_ENV is being overidden by "
                "the configuration option 'env' ({})"
            ).format(options["env"])
        )
    os.environ["FLASK_ENV"] = options["env"]
if "debug" in options:
    if "FLASK_DEBUG" in os.environ and options["debug"] != os.environ["FLASK_DEBUG"]:
        logger.warning(
            (
                "The environment variable FLASK_DEBUG is being overidden by "
                "the configuration option 'debug' ({})"
            ).format(options["debug"])
        )
    os.environ["FLASK_DEBUG"] = options["debug"]

# continue the setup
mail = Mail()
cors = CORS()
compress = Compress()

bootstrap = Bootstrap()

jwt = flask_jwt_extended.JWTManager()
authorize = Authorize(current_user=flask_jwt_extended.get_current_user)
moment = Moment()
toolbar = DebugToolbarExtension()

db = SQLAlchemy()
ma = Marshmallow()


@jwt.user_lookup_loader
def user_loader_callback(jwt_header, jwt_payload):
    """Function for app, to return user object"""

    if jwt_header:
        from seamm_datastore.database.models import User

        username = jwt_payload["sub"]["username"]
        user = User.query.filter_by(username=username).one_or_none()

        return user
    else:
        # return None / null
        return None


def create_app(config_name=None):
    """Flask app factory pattern
    separately creating the extensions and later initializing"""

    conn_app = connexion.App(__name__, specification_dir="./")
    app = conn_app.app

    logger.info("")
    if config_name is not None:
        logger.info("Configuring from configuration " + config_name)
        app.config.from_object(config[config_name])

        options["initialize"] = False
        options["no_check"] = True
    else:
        # Report where options come from
        # parser = configargparse.get_argument_parser("dashboard")
        logger.info("Where options are set:")
        logger.info(60 * "-")
        for section, tmp in parser.get_options().items():
            origin = parser.get_origins(section)
            for key, value in tmp.items():
                logger.info(f"{key:<19} {origin[key]:<15} {value}")

        # Now set the options!
        logger.info("")
        logger.info("Configuration:")
        logger.info(60 * "-")
        for key, value in options.items():
            if key not in (
                "env",
                "debug",
                "initialize",
                "log_dir",
                "log_level",
                "console_log_level",
                "dashboard_configfile",
            ):
                key = key.upper()
                if isinstance(value, str):
                    value = value.replace("%datastore%", datastore)
                logger.info("\t{:>30s} = {}".format(key, value))
                app.config[key] = value

    logger.info("")

    logger.info(
        "Running in "
        + app.config["ENV"]
        + " mode with database "
        + app.config["SQLALCHEMY_DATABASE_URI"]
    )

    # Authorization configuration
    app.config["AUTHORIZE_DEFAULT_PERMISSIONS"] = dict(
        owner=["read", "update", "delete", "create", "manage"],
        group=["read", "update"],
        other=[""],
    )
    app.config["AUTHORIZE_ALLOW_ANONYMOUS_ACTIONS"] = True

    # Set application to store JWTs in cookies.
    app.config["JWT_TOKEN_LOCATION"] = ["cookies", "headers"]

    # Set the cookie paths
    # app.config["JWT_ACCESS_COOKIE_PATH"] = "/api"
    app.config["JWT_REFRESH_COOKIE_PATH"] = "/api/auth/token/refresh"

    # Cookie security
    app.config["JWT_COOKIE_SECURE"] = False
    app.config["JWT_COOKIE_CSRF_PROTECT"] = True
    app.config["JWT_CSRF_ACCESS_PATH"] = "/api/"
    app.config["JWT_ACCESS_TOKEN_EXPIRES"] = timedelta(days=1)
    app.config["JWT_REFRESH_TOKEN_EXPIRES"] = timedelta(days=2)

    conn_app.add_api("swagger.yml")
    db.init_app(app)
    with app.app_context():
        from seamm_datastore.database.build import import_datastore, _build_initial

        if options["initialize"] or config_name and config_name.lower() == "testing":
            logger.warning("Removing all previous jobs from the database.")
            db.drop_all()
            logger.info("Tables dropped creating tables")
            db.create_all()
            # Create database using other interface for consistency.
            logger.info("Importing jobs...")
            _build_initial(db.session, "default")

        if config_name is None or not config_name.lower() == "testing":
            # Log in as user running
            import flask_authorize.plugin
            from seamm_datastore.database.models import User

            flask_authorize.plugin.CURRENT_USER = User.query.filter_by(id=2).one
            logger.warning("Importing any jobs into the database.")
            import_datastore(db.session, str(projects_path))

            flask_authorize.plugin.CURRENT_USER = flask_jwt_extended.get_current_user

        from .routes.auth import auth as auth_blueprint
        from .routes.main import main as main_blueprint
        from .routes.jobs import jobs as jobs_blueprint
        from .routes.flowcharts import flowcharts as flowchart_blueprint
        from .routes.projects import projects as project_blueprint
        from .routes.admin import admin as admin_blueprint

        from .routes.main import errors

        app.register_blueprint(auth_blueprint)
        app.register_blueprint(main_blueprint)
        app.register_blueprint(jobs_blueprint)
        app.register_blueprint(flowchart_blueprint)
        app.register_blueprint(project_blueprint)
        app.register_blueprint(admin_blueprint)

        app.register_error_handler(404, errors.not_found)

    # init
    mail.init_app(app)
    cors.init_app(app)
    bootstrap.init_app(app)
    authorize.init_app(app)
    jwt.init_app(app)
    moment.init_app(app)

    # jinja template
    app.jinja_env.filters["empty"] = replace_empty

    logger.info("")
    logger.info("Final configuration:")
    logger.info(60 * "-")
    for key, value in app.config.items():
        logger.info("\t{:>30s} = {}".format(key, value))
    logger.info("")

    logger.info(f"{app.url_map}")

    # apply compress
    compress.init_app(app)

    return app



from flask_wtf import FlaskForm
from flask_wtf.file import FileField
from wtforms import StringField, TextAreaField, SubmitField

from wtforms.validators import DataRequired, Length, Regexp


class EditJob(FlaskForm):
    name = StringField(
        "Job Title",
        validators=[
            DataRequired(),
            Length(1, 64),
            Regexp(
                "^[A-Za-z][A-Za-z0-9_. ]*$",
                0,
                "Job titles must have only letters, numbers, "
                + "dots, spaces, or underscores",
            ),
        ],
    )
    notes = TextAreaField("Description")
    submit = SubmitField("Update Job")


class ImportJob(FlaskForm):
    outfile = FileField("job_data.json File", validators=[DataRequired()])
    title = StringField("Job Title")
    submit = SubmitField("Import Job")



from .jobs import *  # noqa: F401, F403
from .flowcharts import *  # noqa: F401, F403
from .projects import *  # noqa: F401, F403
from .status import *  # noqa: F401, F403
from .users import *  # noqa: F401, F403
from .groups import get_groups  # noqa: F401, F403
from .roles import get_roles  # noqa: F401, F403
from .auth import *  # noqa: F401, F403



import logging

from flask import render_template, flash, make_response, request, redirect, url_for

from flask_jwt_extended import (
    set_access_cookies,
    set_refresh_cookies,
    unset_jwt_cookies,
    create_access_token,
    get_current_user,
    jwt_required,
)

from .forms import LoginForm, ConfirmLogin, UpdateAccountInfoForm

from seamm_dashboard import authorize, db
from seamm_datastore.database.models import User, UserProjectAssociation
from seamm_datastore.database.schema import UserSchema

from . import auth

from seamm_dashboard import jwt
from seamm_dashboard.routes.api.auth import create_tokens, refresh_expiring_jwts

from seamm_dashboard.routes.admin.forms import _validate_email
from seamm_dashboard.routes.admin.views import _bind_user_projects_to_form

logger = logging.getLogger(__name__)


@auth.route("/login", methods=["GET", "POST"])
def login():

    form = LoginForm()
    if form.validate_on_submit():
        user = User.query.filter_by(username=form.username.data).one_or_none()

        if user is not None and user.verify_password(form.password.data):
            # redirect to blank login page which will set local storage to
            # reload all tabs (as logged in user)
            response = make_response(render_template("login_script.html"))

            # Add cookies to response
            access_token, refresh_token = create_tokens(user)
            set_access_cookies(response, access_token)
            set_refresh_cookies(response, refresh_token)

            flash(f"You have been logged in as {user.username}.", "success")

            return response
        flash("Invalid username or password.")

    return render_template("auth/login.html", form=form)


@jwt.needs_fresh_token_loader
@jwt_required(optional=True)
@auth.route("/confirm_login", methods=["GET", "POST"])
def fresh_login(jwt_header, jwt_payload):

    form = ConfirmLogin()
    user = get_current_user()

    if form.validate_on_submit():

        if user.verify_password(form.password.data):
            user_schema = UserSchema(many=False)
            user = user_schema.dump(user)
            response = redirect(request.referrer)

            # Add cookies to response
            access_token = create_access_token(identity=user, fresh=True)
            set_access_cookies(response, access_token)

            return response
        flash("Invalid username or password.")

    return render_template("auth/confirm_login.html", form=form)


@auth.route("/my-account", methods=["GET", "POST"])
@jwt_required(optional=True, fresh=True)
def my_account():

    current_user = get_current_user()

    # Redirect to admin version if logged in user is admin
    if authorize.has_role("admin"):
        return redirect(url_for("admin.manage_user", user_id=current_user.id))

    if current_user is None:
        return redirect(url_for("auth.login"))

    project_association = UserProjectAssociation.query.filter_by(
        entity_id=current_user.id
    ).all()

    projects = []
    if project_association:
        projects = [x.projects for x in project_association]

    form = UpdateAccountInfoForm
    form, project_names = _bind_user_projects_to_form(
        form, projects, user=current_user, new=False
    )

    form = form()

    if request.method == "GET":
        form.first_name.data = current_user.first_name
        form.last_name.data = current_user.last_name
        form.email.data = current_user.email

    if request.method == "POST":

        try:
            form.email.validators.remove(_validate_email)
        except ValueError:
            pass

        if form.validate():
            current_user.first_name = form.first_name.data
            current_user.last_name = form.last_name.data
            current_user.email = form.email.data

            if form.password:
                current_user.password = form.password.data

            db.session.add(current_user)
            db.session.commit()

            flash("Your account has been updated.")

            return redirect(url_for("main.index"))

    return render_template(
        "auth/manage_account.html", form=form, user=current_user, projects=project_names
    )


# Login required.
@auth.route("/logout")
def logout():
    """
    Direct to blank page which sets local storage to log out all other tabs and
    redirects to main
    """

    from flask import session

    if "expired=True" in request.environ.get("RAW_URI", []):
        flash("You have been logged out due to inactivity.")
    else:
        flash("You have been logged out.")
    response = make_response(render_template("logout.html"))
    unset_jwt_cookies(response)

    flashes = session.get("_flashes", [])

    if len(flashes) > 1:
        flashes = [x for x in flashes if "inactivity" in flashes[1]]

    session["_flashes"] = flashes

    return response


@auth.after_request
def auth_refresh(response):
    return refresh_expiring_jwts(response)



from copy import deepcopy
import os

from flask import render_template, url_for, request, flash, redirect
from flask_jwt_extended import jwt_required, get_current_user

from sqlalchemy import func

from pathlib import Path

from wtforms import BooleanField

from . import projects
from .forms import EditProject, ManageProjectAccessForm, AddProject

from seamm_datastore.database.models import (
    Project,
    User,
    UserProjectAssociation,
    Job,
)  # noqa: E501

from seamm_dashboard import authorize, db, datastore

from seamm_dashboard.routes.api.auth import refresh_expiring_jwts


def _bind_users_to_form(form, current_user, project_id):

    actions = ["read", "update", "create", "delete", "manage"]
    users = User.query.all()
    user_names = []

    if len(users) > 1:
        # Put current user first
        users.remove(current_user)
        reorder_users = [current_user]
        reorder_users.extend(users)
        users = reorder_users

    for user in users:
        field_name = f"user_{user.id}"
        permissions = []
        assoc = UserProjectAssociation.query.filter_by(
            resource_id=project_id, entity_id=user.id
        ).one_or_none()

        if assoc:
            permissions = assoc.permissions

        for action in actions:
            checked = action in permissions
            setattr(form, f"{field_name}_{action}", BooleanField(default=checked))

        user_names.append({"username": user.username, "id": user.id})

    return form, user_names


@projects.route("/views/projects")
@jwt_required(optional=True)
def project_list():
    return render_template("projects/project_list.html")


@projects.route("/views/projects/<id>/jobs")
@projects.route("/views//projects/<id>/jobs")
@jwt_required(optional=True)
def project_jobs_list(id):

    project = Project.query.get(id)

    manage_project = authorize.manage(project)
    edit_project = authorize.update(project)

    # Build the url ourselves.
    base_url = url_for("main.index")
    edit_url = base_url + f"projects/{id}/edit"
    manage_url = base_url + f"projects/{id}/manage"

    return render_template(
        "jobs/jobs_list.html",
        project=True,
        manage_project=manage_project,
        edit_project=edit_project,
        edit_url=edit_url,
        manage_url=manage_url,
    )


@projects.route("/projects/<project_id>/edit", methods=["GET", "POST"])
@jwt_required(optional=True)
def edit_project(project_id):

    project = Project.query.get(project_id)

    if not authorize.update(project):
        return render_template("401.html")

    form = EditProject()

    # Build the url ourselves.
    base_url = url_for("main.index")
    project_url = base_url + f"#projects/{project_id}/jobs"

    if form.validate_on_submit():
        # Rename project path
        path = project.path
        new_path = os.path.join(os.path.dirname(path), form.name.data)

        # Make sure project name doesn't exist
        exists = Project.query.filter(Project.name == form.name.data).one_or_none()
        if exists is not None:
            flash(f"A project with the name {form.name.data} already exists.")
            return redirect(url_for("projects.edit_project", project_id=project_id))

        # Need to change this to query, but this will work
        # without modifying the DB for now.
        job_ids = [x.id for x in project.jobs]

        # Update job paths
        Job.query.filter(Job.id.in_(job_ids)).update(
            {Job.path: func.replace(Job.path, path, new_path)},
            synchronize_session=False,
        )

        # Perform rename
        os.rename(path, new_path)

        # Update project information
        project.path = new_path
        project.name = form.name.data
        project.description = form.notes.data
        db.session.commit()
        flash("Project updated successfully.", "successs")

        return redirect(project_url)

    elif request.method == "GET":
        form.name.data = project.name
        form.notes.data = project.description

    return render_template(
        "jobs/edit_job.html",
        title=f"Edit Project {project_id}",
        form=form,
        back_url=project_url,
    )


@projects.route("/projects/<project_id>/manage", methods=["GET", "POST"])
@jwt_required(optional=True)
def manage_project(project_id):

    project = Project.query.get(project_id)

    if not project:
        return render_template("404.html")

    if not authorize.manage(project):
        return render_template("401.html")

    form = deepcopy(ManageProjectAccessForm)

    form, usernames = _bind_users_to_form(
        form, current_user=get_current_user(), project_id=project.id
    )

    form = form()

    # Build the url ourselves.
    base_url = url_for("main.index")
    project_url = base_url + f"#projects/{project_id}/jobs"

    if request.method == "POST":
        if form.validate_on_submit():

            user_keys = [
                x for x in form.data.keys() if "user" in x if form.data[x] is True
            ]

            permissions_dict = {}

            for key in user_keys:
                split = key.split("_")
                user_id = int(split[1])
                permission = split[2]

                try:
                    permissions_dict[user_id].append(permission)
                except KeyError:
                    permissions_dict[user_id] = [permission]

            # Find an entries for special user permissions which exist for this project
            users = User.query.all()

            for entry in users:
                user_id = entry.id

                try:
                    permissions = permissions_dict[user_id]
                except KeyError:
                    permissions = []

                assoc = UserProjectAssociation.query.filter_by(
                    entity_id=user_id, resource_id=project_id
                ).one_or_none()

                if not assoc:
                    assoc = UserProjectAssociation(
                        entity_id=user_id,
                        resource_id=project_id,
                        permissions=permissions,
                    )
                else:
                    assoc.permissions = permissions

                db.session.add(assoc)
                db.session.commit()

            flash(f"Permissions for {project.name} successfully updated.")
            return redirect(project_url)

    return render_template(
        "projects/project_access.html",
        form=form,
        users=usernames,
        project=project,
        back_url=project_url,
    )


@projects.route("/projects/add", methods=["GET", "POST"])
@jwt_required(optional=True)
def add_project():

    if get_current_user() is None:
        return render_template("404.html")

    form = AddProject()

    # Build the url ourselves.
    base_url = url_for("main.index")
    project_url = base_url + "#projects"

    if form.validate_on_submit():
        # Create a directory for the project
        project_path = Path(datastore).expanduser() / "projects" / form.name.data

        project_path.mkdir(parents=True, exist_ok=True)
        project = Project.create(
            name=form.name.data, description=form.notes.data, path=str(project_path)
        )
        db.session.add(project)
        db.session.commit()
        flash(f"Project {project.name} added.", "successs")

        return redirect(project_url)

    return render_template(
        "jobs/edit_job.html",
        title="Add project",
        form=form,
        back_url=project_url,
    )


@projects.after_request
def project_refresh(response):
    return refresh_expiring_jwts(response)



from flask_wtf import FlaskForm
from wtforms import StringField, TextAreaField, SubmitField

from wtforms.validators import DataRequired, Length, Regexp


class EditProject(FlaskForm):
    name = StringField(
        "Project Title",
        validators=[
            DataRequired(),
            Length(1, 64),
            Regexp(
                "^[A-Za-z][A-Za-z0-9_. ]*$",
                0,
                "Projects must have only letters, numbers, "
                + "dots, spaces, or underscores",
            ),
        ],
    )
    notes = TextAreaField("Description")
    submit = SubmitField("Update Project")


class AddProject(FlaskForm):
    name = StringField(
        "Project Title",
        validators=[
            DataRequired(),
            Length(1, 64),
            Regexp(
                "^[A-Za-z][A-Za-z0-9_. ]*$",
                0,
                "Projects must have only letters, numbers, "
                + "dots, spaces, or underscores",
            ),
        ],
    )
    notes = TextAreaField("Description")
    submit = SubmitField("Create Project")


class ManageProjectAccessForm(FlaskForm):

    submit = SubmitField("Update Project Access")
